### Y: data matrix of size n_s*(\sum_i^{n_p}n_vi); row: site, column: visit for a subject  
### Q: basis function matrix of size n_s*L, L is no. of eigenvectors of Cov(Y)
### Time: sum(nv)*1 vector using the first visit=0 as baseline
### time.ind: the index of first visit for each subject, np*1 vector
### omega: a vector of eigenvalues of Cov(Y), L*1 vector
### np: no. of subjects
### nv: no. of visits for each subject, np*1 vector 
### ns: no. of sites
###  a: baseline PD, L*np matrix
###  e: spatial effect, L*np matrix
###  d: shrinkage, L*np matrix
###  z: normal rv, L*np matrix
### gm: scale parameter in variance of z, np*1 vector

###Let sigma_a_i be sigma_a and gamma_i be gamma (i.e. same for all subjects)

sths0_new <- function(Y.obs, np, Time, Q, omega, miss.sub, iters=5000, burn=1000, update=1000){
  library(MCMCpack)
  library(mvtnorm)
  library(FastGP)
  
  bgn.ind <- which(Time==0)
  end.ind <- c(bgn.ind[-1]-1,ncol(Y.obs))
  miss    <- end.ind[miss.sub]
  n.miss  <- length(miss.sub) 
  
  ns <- nrow(Y.obs)
  Nv <- ncol(Y.obs)
  nv <- end.ind-bgn.ind+1
  L  <- ncol(Q)
  
  Y        <- Y.obs
  Y[,miss] <- Y.obs[,(miss-1)]
  Y        <- matrix(runif(ns*Nv,Y-.5,Y+.5),ns,Nv)
  
  #standardize Q
  Q   <- sweep(Q,1,sqrt(rowSums(Q^2)),"/")
  Q   <- Q+0.001
  tQ  <- t(Q) 
  tQQ <- crossprod(Q)
  
  inv.Omega <- diag(1/omega,length(omega))
  w <- sqrt(diag(Q%*%diag(omega)%*%t(Q)))
  
  #initial values
  gm2 <- sigma_a2 <- 1
  a   <- z <- mu_a <- matrix(0.1,L,np)
  inv.Sigma_a <- inv.Sigma_z <- diag(L)
  sigma_eps2 <- 1
  
  Qa <- Q%*%a
  Qz <- Q%*%z
  
  mu <- matrix(0,ns,Nv)
  beta <- matrix(0,ns,np)
  for (j in 1:np){  
    beta[,j] <- Qz[,j]
    mu[,(bgn.ind[j]:end.ind[j])] <- Qa[,j]+outer(beta[,j],Time[(bgn.ind[j]:end.ind[j])])
  }
  
  #Set up storage
  B1 <- B2 <- B3 <- matrix(0,nrow=ns,ncol=np)
  M1 <- M2 <- matrix(0,nrow=ns,ncol=Nv)
  P1 <- P2 <- matrix(0,nrow=ns,ncol=n.miss)
  Sa <- Sz <- matrix(0,nrow=L,ncol=L)
  Y.pred <- matrix(0,nrow=ns,ncol=n.miss)
  
  #MH 
  att <- acc <- rep(0,np)
  MH  <- rep(1,np)
  
  tick <- proc.time()[3]
  
  #Start MCMC
  for (iter in 1:iters){
    sum.outer_a <- sum.outer_z <- matrix(0,L,L)
    
    #Update predicted Y
    Y.pred       <- matrix(rnorm(ns*n.miss,mu[,miss],sqrt(sigma_eps2)),ns,n.miss)
    Y.obs[,miss] <- round(Y.pred)
    
    for (sub in 1:np){
      ind     <- bgn.ind[sub]:end.ind[sub]
      Yb.diff <- Y[,ind]-outer(beta[,sub],Time[ind])
      
      #Update Y_ij
      Y[,ind] <- matrix(rtnorm(ns*length(ind),Y.obs[,ind]-.5,Y.obs[,ind]+.5,mu[,ind],sqrt(sigma_eps2)),ns,length(ind))
      Y[,ind] <- ifelse(is.infinite(Y[,ind]),yes=Y.obs[,ind],no=Y[,ind])
      
      #Update a_i
      post.cov_a <- chol2inv(chol(inv.Sigma_a/sigma_a2 + (nv[sub]/sigma_eps2)*tQQ))
      post.mn_a  <- (1/sigma_eps2)*post.cov_a%*%tQ%*%rowSums(Yb.diff) + (1/sigma_a2)*post.cov_a%*%inv.Sigma_a%*%mu_a[,sub]
      a[,sub]    <- as.vector(rcpp_rmvnorm(1,post.cov_a,post.mn_a))
      Qa[,sub]   <- Q%*%a[,sub] 
      
      #Update mu_a_i
      post.cov_mua <- chol2inv(chol(inv.Sigma_a/sigma_a2 + diag(1/10,L)))
      post.mn_mua  <- post.cov_mua%*%(inv.Sigma_a/sigma_a2)%*%a[,sub]
      mu_a[,sub]   <- as.vector(rcpp_rmvnorm(1,post.cov_mua,post.mn_mua))
      
      #Update z_i
      Cur_z        <- .5*nlp(Y[,ind],mu[,ind],Qa[,sub],rep(0,ns),Qz[,sub],Time[ind])/sigma_eps2
      att[sub]     <- att[sub]+1
      Cov_z        <- MH[sub]*chol2inv(chol(inv.Sigma_z))*gm2
      can_z        <- as.vector(z[,sub]+rmvnorm(1,mean=rep(0,L),sigma=Cov_z)) 
      Qcan_z       <- Q%*%can_z 
      Can_z        <- .5*nlp(Y[,ind],mu[,ind],Qa[,sub],rep(0,ns),as.vector(Qcan_z),Time[ind])/sigma_eps2
      rz           <- Cur_z-Can_z+dmvnorm(can_z,mean=rep(0,L),sigma=Cov_z/MH[sub],log=TRUE)-dmvnorm(z[,sub],mean=rep(0,L),sigma=Cov_z/MH[sub],log=TRUE)
      if(log(runif(1))<rz){
        acc[sub]   <- acc[sub]+1
        z[,sub]    <- can_z
        Cur_z      <- Can_z
      }    
      Qz[,sub]     <- Q%*%z[,sub]
      
      #Calculate beta and mu
      beta[,sub]  <- as.vector(Qz[,sub])
      tBeta       <- outer(beta[,sub],Time[ind])
      mu[,ind]    <- Qa[,sub]+tBeta
      
      #Calculate sum of outer product for Wishart posterior of Sigma_a, z
      sum.outer_a <- sum.outer_a+(1/sigma_a2)*outer(a[,sub],a[,sub])
      sum.outer_z <- sum.outer_z+(1/gm2)*outer(z[,sub],z[,sub])
    }
    
    #Update gm, sigma_a2
    SS_gm2        <- sum((sum.outer_z*gm2)*inv.Sigma_z)
    gm2           <- 1/rgamma(1,.1+.5*np,.1+.5*SS_gm2)
    
    SS_a2         <- sum((sum.outer_a*sigma_a2)*inv.Sigma_a)
    sigma_a2      <- 1/rgamma(1,.1+.5*np,.1+.5*SS_a2)
    
    #Update scale matrices
    inv.Sigma_a <- rwish(L+np,chol2inv(chol(sum.outer_a+inv.Omega)))
    inv.Sigma_z <- rwish(L+np,chol2inv(chol(sum.outer_z+inv.Omega)))
    
    #Update sigma_eps2
    SS_eps2    <- sum((Y-mu)^2)
    sigma_eps2 <- 1/rgamma(1,.1+.5*sum(nv)*ns,.1+.5*SS_eps2)
    
    #Metropolis tuning
    for(j in 1:length(att)){if(iter<burn & att[j]>25){
      if(acc[j]/att[j]<0.3){MH[j] <- MH[j]*0.9}
      if(acc[j]/att[j]>0.5){MH[j] <- MH[j]*1.1}
      acc[j] <- att[j] <-0
    }}
    
    if(iter>burn){
      B1 <- B1 + beta/(iters-burn)
      B2 <- B2 + beta*beta/(iters-burn)
      B3 <- B3 + (beta>0)/(iters-burn)
      M1 <- M1 + mu/(iters-burn)
      M2 <- M2 + mu*mu/(iters-burn)
      P1 <- P1 + Y.pred/(iters-burn)
      P2 <- P2 + Y.pred*Y.pred/(iters-burn)
      Sa <- Sa + chol2inv(inv.Sigma_a)/(iters-burn)
      Sz <- Sz + chol2inv(inv.Sigma_z)/(iters-burn)
    }
    
#    if(iter%%update==0){
#      print(paste("Done with",iter,"of",iters))
#    }
  }
  
  tock <- proc.time()[3]
  out <-list(bt.mn=B1,bt.sd=sqrt(B2-B1^2),bt.pro=B3,
             mu.mn=M1,mu.sd=sqrt(M2-M1^2),
             pred.mn=P1,pred.sd=sqrt(P2-P1^2),
             Sigma.a=Sa,Sigma.z=Sz,
             acc.rate=acc/att,minutes=(tock-tick)/60)
  return(out)
}


nlp <- function(Y,mu,Qa,Qd,Qz,Time){
  b      <- tan(.5*pi*pnorm(Qd))*(Qz)
  tb     <- outer(b,Time) 
  mu     <- Qa+tb
  out    <- sum((Y-mu)^2)
  return(out)  
}

rtnorm <- function(n,l,u,m,s){
  lo <- pnorm(l,m,s)
  hi <- pnorm(u,m,s)
  Y  <- runif(n,lo,hi)
  Y  <- qnorm(Y,m,s)
  return(Y)
}

